@@ -1,4 +1,5 @@
 /* See LICENSE file for copyright and license details. */
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
@@ -27,6 +28,19 @@ static int devnullfd;
 static int parent = 1;
 static int interactive;
 
+static void (*diag)(char *fmt, ...);
+
+void
+stddiag(char *fmt, ...)
+{
+	va_list arg;
+
+	va_start(arg, fmt);
+	vfprintf(stderr, fmt, arg);
+	va_end(arg);
+	fputc('\n', stderr);
+}
+
 void
 die(const char *fmt, ...)
 {
@@ -40,7 +54,7 @@ die(const char *fmt, ...)
 	exit(1);
 }
 
-#ifndef asprintf
+#ifdef NEED_ASPRINTF
 int
 asprintf(char **s, const char *fmt, ...)
 {
@@ -60,7 +74,28 @@ asprintf(char **s, const char *fmt, ...)
 
 	return n;
 }
-#endif /* asprintf */
+#endif /* NEED_ASPRINTF */
+
+#ifdef NEED_STRCASESTR
+char *
+strcasestr(const char *h, const char *n)
+{
+	size_t i;
+
+	if (!n[0])
+		return (char *)h;
+
+	for (; *h; ++h) {
+		for (i = 0; n[i] && tolower((unsigned char)n[i]) ==
+		            tolower((unsigned char)h[i]); ++i)
+			;
+		if (n[i] == '\0')
+			return (char *)h;
+	}
+
+	return NULL;
+}
+#endif /* NEED_STRCASESTR */
 
 /* print `len' columns of characters. */
 size_t
@@ -113,7 +148,7 @@ xmalloc(const size_t n)
 	void *m = malloc(n);
 
 	if (!m)
-		die("malloc: %s\n", strerror(errno));
+		die("malloc: %s", strerror(errno));
 
 	return m;
 }
@@ -121,10 +156,10 @@ xmalloc(const size_t n)
 static void *
 xcalloc(size_t n)
 {
-	char *m = xmalloc(n);
+	char *m = calloc(1, n);
 
-	while (n)
-		m[--n] = 0;
+	if (!m)
+		die("calloc: %s", strerror(errno));
 
 	return m;
 }
@@ -135,7 +170,7 @@ xstrdup(const char *str)
 	char *s;
 
 	if (!(s = strdup(str)))
-		die("strdup: %s\n", strerror(errno));
+		die("strdup: %s", strerror(errno));
 
 	return s;
 }
@@ -194,7 +229,7 @@ typedisplay(char t)
 	case '7':
 		return "Find+";
 	case '8':
-		return "Tlnt|";
+		return "Tlnt+";
 	case '9':
 		return "Binf+";
 	case '+':
@@ -209,10 +244,12 @@ typedisplay(char t)
 		return "HTML+";
 	case 'i':
 		return "    |";
-	case 's':
-		return "Snd |";
 	default:
-		return "!   |";
+		/* "Characters '0' through 'Z' are reserved." (ASCII) */
+		if (t >= '0' && t <= 'Z')
+			return "!   |";
+		else
+			return "UNKN|";
 	}
 }
 
@@ -244,11 +281,12 @@ displaytextitem(Item *item)
 	uicleanup();
 	switch (pid = fork()) {
 	case -1:
-		uistatus("Couldn't fork.");
+		diag("Couldn't fork.");
 		return;
 	case 0:
 		parent = 0;
-		pagerin = popen("$PAGER", "we");
+		if (!(pagerin = popen("$PAGER", "we")))
+			_exit(1);
 		fputs(item->raw, pagerin);
 		exit(pclose(pagerin));
 	default:
@@ -259,15 +297,18 @@ displaytextitem(Item *item)
 }
 
 static char *
-pickfield(char **raw, char sep)
+pickfield(char **raw, const char *sep)
 {
-	char *c, *f = *raw;
+	char c, *r, *f = *raw;
 
-	for (c = *raw; *c && *c != sep; ++c)
-		;
+	for (r = *raw; (c = *r) && !strchr(sep, c); ++r) {
+		if (c == '\n')
+			goto skipsep;
+	}
 
-	*c = '\0';
-	*raw = c+1;
+	*r++ = '\0';
+skipsep:
+	*raw = r;
 
 	return f;
 }
@@ -282,10 +323,12 @@ invaliditem(char *raw)
 		if (c == '\t')
 			++tabs;
 	}
-	if (c)
+	if (tabs < 3) {
 		*raw++ = '\0';
+		return raw;
+	}
 
-	return (tabs == 3) ? NULL : raw;
+	return NULL;
 }
 
 static void
@@ -303,21 +346,22 @@ molditem(Item *item, char **raw)
 	}
 
 	item->type = *raw[0]++;
-	item->username = pickfield(raw, '\t');
-	item->selector = pickfield(raw, '\t');
-	item->host = pickfield(raw, '\t');
-	item->port = pickfield(raw, '\r');
-	if (!*raw[0])
+	item->username = pickfield(raw, "\t");
+	item->selector = pickfield(raw, "\t");
+	item->host = pickfield(raw, "\t");
+	item->port = pickfield(raw, "\t\r");
+	while (*raw[0] != '\n')
 		++*raw;
+	*raw[0]++ = '\0';
 }
 
 static Dir *
 molddiritem(char *raw)
 {
-	Item *items = NULL;
+	Item *item, *items = NULL;
 	char *s, *nl, *p;
 	Dir *dir;
-	size_t i, nitems;
+	size_t i, n, nitems;
 
 	for (s = nl = raw, nitems = 0; p = strchr(nl, '\n'); ++nitems) {
 		s = nl;
@@ -326,7 +370,7 @@ molddiritem(char *raw)
 	if (!strcmp(s, ".\r\n") || !strcmp(s, ".\n"))
 		--nitems;
 	if (!nitems) {
-		uistatus("Couldn't parse dir item");
+		diag("Couldn't parse dir item");
 		return NULL;
 	}
 
@@ -334,8 +378,18 @@ molddiritem(char *raw)
 	items = xreallocarray(items, nitems, sizeof(Item));
 	memset(items, 0, nitems * sizeof(Item));
 
-	for (i = 0; i < nitems; ++i)
-		molditem(&items[i], &raw);
+	for (i = 0; i < nitems; ++i) {
+		item = &items[i];
+		molditem(item, &raw);
+		if (item->type == '+') {
+			for (n = i - 1; n < (size_t)-1; --n) {
+				if (items[n].type != '+') {
+					item->redtype = items[n].type;
+					break;
+				}
+			}
+		}
+	}
 
 	dir->items = items;
 	dir->nitems = nitems;
@@ -367,7 +421,7 @@ getrawitem(int sock)
 	*buf = '\0';
 
 	if (n < 0) {
-		uistatus("Can't read socket: %s", strerror(errno));
+		diag("Can't read socket: %s", strerror(errno));
 		clear(&raw);
 	}
 
@@ -385,14 +439,14 @@ sendselector(int sock, const char *selector)
 	msg = p = xmalloc(ln);
 	snprintf(msg, ln--, "%s\r\n", selector);
 
-	while ((n = write(sock, p, ln)) != -1 && n != 0) {
+	while ((n = write(sock, p, ln)) > 0) {
 		ln -= n;
 		p += n;
 	}
 
 	free(msg);
 	if (n == -1)
-		uistatus("Can't send message: %s", strerror(errno));
+		diag("Can't send message: %s", strerror(errno));
 
 	return n;
 }
@@ -400,18 +454,23 @@ sendselector(int sock, const char *selector)
 static int
 connectto(const char *host, const char *port)
 {
+	sigset_t set, oset;
 	static const struct addrinfo hints = {
 	    .ai_family = AF_UNSPEC,
 	    .ai_socktype = SOCK_STREAM,
 	    .ai_protocol = IPPROTO_TCP,
 	};
 	struct addrinfo *addrs, *addr;
-	int sock, r;
+	int r, sock = -1;
+
+	sigemptyset(&set);
+	sigaddset(&set, SIGWINCH);
+	sigprocmask(SIG_BLOCK, &set, &oset);
 
 	if (r = getaddrinfo(host, port, &hints, &addrs)) {
-		uistatus("Can't resolve hostname \"%s\": %s",
-		         host, gai_strerror(r));
-		return -1;
+		diag("Can't resolve hostname \"%s\": %s",
+		     host, gai_strerror(r));
+		goto err;
 	}
 
 	for (addr = addrs; addr; addr = addr->ai_next) {
@@ -424,19 +483,25 @@ connectto(const char *host, const char *port)
 		}
 		break;
 	}
-	if (sock < 0) {
-		uistatus("Can't open socket: %s", strerror(errno));
-		return -1;
-	}
-	if (r < 0) {
-		uistatus("Can't connect to: %s:%s: %s",
-		         host, port, strerror(errno));
-		return -1;
-	}
 
 	freeaddrinfo(addrs);
 
+	if (sock < 0) {
+		diag("Can't open socket: %s", strerror(errno));
+		goto err;
+	}
+	if (r < 0) {
+		diag("Can't connect to: %s:%s: %s",
+		     host, port, strerror(errno));
+		goto err;
+	}
+
+	sigprocmask(SIG_SETMASK, &oset, NULL);
 	return sock;
+
+err:
+	sigprocmask(SIG_SETMASK, &oset, NULL);
+	return -1;
 }
 
 static int
@@ -457,6 +522,7 @@ download(Item *item, int dest)
 		return 0;
 	}
 
+	w = 0;
 	while ((r = read(src, buf, BUFSIZ)) > 0) {
 		while ((w = write(dest, buf, r)) > 0)
 			r -= w;
@@ -500,8 +566,8 @@ downloaditem(Item *item)
 	}
 
 	if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) < 0) {
-		uistatus("Can't open destination file %s: %s",
-		         path, strerror(errno));
+		diag("Can't open destination file %s: %s",
+		     path, strerror(errno));
 		errno = 0;
 		goto cleanup;
 	}
@@ -509,8 +575,11 @@ downloaditem(Item *item)
 	if (!download(item, dest))
 		goto cleanup;
 
-	if (!item->tag)
-		item->tag = path;
+	if (item->tag)
+		goto cleanup;
+
+	item->tag = path;
+
 	return;
 cleanup:
 	free(path);
@@ -529,7 +598,7 @@ fetchitem(Item *item)
 	close(sock);
 
 	if (item->raw && !*item->raw) {
-		uistatus("Empty response from server");
+		diag("Empty response from server");
 		clear(&item->raw);
 	}
 
@@ -541,17 +610,17 @@ plumb(char *url)
 {
 	switch (fork()) {
 	case -1:
-		uistatus("Couldn't fork.");
+		diag("Couldn't fork.");
 		return;
 	case 0:
 		parent = 0;
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
 		if (execlp(plumber, plumber, url, NULL) < 0)
-			uistatus("execlp: plumb(%s): %s", url, strerror(errno));
+			_exit(1);
 	}
 
-	uistatus("Plumbed \"%s\"", url);
+	diag("Plumbed \"%s\"", url);
 }
 
 static void
@@ -559,7 +628,7 @@ plumbitem(Item *item)
 {
 	char *file, *path, *tag;
 	mode_t mode = S_IRUSR|S_IWUSR|S_IRGRP;
-	int n, dest, plumbitem;
+	int dest, plumbitem;
 
 	if (file = strrchr(item->selector, '/'))
 		++file;
@@ -582,15 +651,15 @@ plumbitem(Item *item)
 		clear(&path);
 		if (!tag) {
 			if (asprintf(&path, "%s/%s", tmpdir, file) < 0)
-				die("Can't generate tmpdir path: ",
-				    strerror(errno));
+				die("Can't generate tmpdir path: %s/%s: %s",
+				    tmpdir, file, strerror(errno));
 		}
 	}
 
 	if (path && (!tag || strcmp(tag, path))) {
 		if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) < 0) {
-			uistatus("Can't open destination file %s: %s",
-			         path, strerror(errno));
+			diag("Can't open destination file %s: %s",
+			     path, strerror(errno));
 			errno = 0;
 			goto cleanup;
 		}
@@ -614,13 +683,15 @@ static int
 dig(Item *entry, Item *item)
 {
 	char *plumburi = NULL;
+	int t;
 
 	if (item->raw) /* already in cache */
 		return item->type;
 	if (!item->entry)
 		item->entry = entry ? entry : item;
 
-	switch (item->type) {
+	t = item->redtype ? item->redtype : item->type;
+	switch (t) {
 	case 'h': /* fallthrough */
 		if (!strncmp(item->selector, "URL:", 4)) {
 			plumb(item->selector+4);
@@ -631,7 +702,6 @@ dig(Item *entry, Item *item)
 			return 0;
 		break;
 	case '1':
-	case '+':
 	case '7':
 		if (!fetchitem(item) || !(item->dat = molddiritem(item->raw)))
 			return 0;
@@ -643,26 +713,30 @@ dig(Item *entry, Item *item)
 		downloaditem(item);
 		return 0;
 	case '8':
-		if (asprintf(&plumburi, "telnet://%s@%s:%s", item->selector,
+		if (asprintf(&plumburi, "telnet://%s%s%s:%s",
+		             item->selector, item->selector ? "@" : "",
 		             item->host, item->port) < 0)
 			return 0;
 		plumb(plumburi);
 		free(plumburi);
 		return 0;
-	case 'g':
-	case 'I':
-		plumbitem(item);
-		return 0;
 	case 'T':
-		if (asprintf(&plumburi, "tn3270://%s@%s:%s", item->selector,
+		if (asprintf(&plumburi, "tn3270://%s%s%s:%s",
+		             item->selector, item->selector ? "@" : "",
 		             item->host, item->port) < 0)
 			return 0;
 		plumb(plumburi);
 		free(plumburi);
 		return 0;
 	default:
-		uistatus("Type %c (%s) not supported",
-		        item->type, typedisplay(item->type));
+		if (t >= '0' && t <= 'Z') {
+			diag("Type %c (%s) not supported", t, typedisplay(t));
+			return 0;
+		}
+	case 'g':
+	case 'I':
+		plumbitem(item);
+	case 'i':
 		return 0;
 	}
 
@@ -701,8 +775,9 @@ searchitem(Item *entry, Item *item)
 	if (!(sel = searchselector(item)))
 		return 0;
 
-	if (sel != item->tag) {
+	if (sel != item->tag)
 		clearitem(item);
+	if (!item->dat) {
 		selector = item->selector;
 		item->selector = item->tag = sel;
 		dig(entry, item);
@@ -714,19 +789,37 @@ searchitem(Item *entry, Item *item)
 static void
 printout(Item *hole)
 {
+	char t = 0;
+
 	if (!hole)
 		return;
 
-	switch (hole->type) {
+	switch (hole->redtype ? hole->redtype : (t = hole->type)) {
 	case '0':
 		if (dig(hole, hole))
 			fputs(hole->raw, stdout);
 		return;
 	case '1':
-	case '+':
+	case '7':
 		if (dig(hole, hole))
 			printdir(hole);
+		return;
 	default:
+		if (t >= '0' && t <= 'Z') {
+			diag("Type %c (%s) not supported", t, typedisplay(t));
+			return;
+		}
+	case '4':
+	case '5':
+	case '6':
+	case '9':
+	case 'g':
+	case 'I':
+		download(hole, 1);
+	case '2':
+	case '3':
+	case '8':
+	case 'T':
 		return;
 	}
 }
@@ -737,7 +830,7 @@ delve(Item *hole)
 	Item *entry = NULL;
 
 	while (hole) {
-		switch (hole->type) {
+		switch (hole->redtype ? hole->redtype : hole->type) {
 		case 'h':
 		case '0':
 			if (dig(entry, hole))
@@ -752,6 +845,10 @@ delve(Item *hole)
 			if (searchitem(entry, hole))
 				entry = hole;
 			break;
+		case 0:
+			diag("Couldn't get %s:%s/%c%s", hole->host,
+			     hole->port, hole->type, hole->selector);
+			break;
 		case '4':
 		case '5':
 		case '6': /* TODO decode? */
@@ -760,11 +857,9 @@ delve(Item *hole)
 		case 'g':
 		case 'I':
 		case 'T':
+		default:
 			dig(entry, hole);
 			break;
-		case 0:
-			uistatus("Couldn't get %s:%s/%c%s", hole->host,
-			         hole->port, hole->type, hole->selector);
 		}
 
 		if (!entry)
@@ -827,6 +922,16 @@ moldentry(char *url)
 	entry = xcalloc(sizeof(Item));
 	entry->type = gopherpath[0];
 	entry->username = entry->selector = ++gopherpath;
+	if (entry->type == '7') {
+		if (p = strstr(gopherpath, "%09")) {
+			memmove(p+1, p+3, strlen(p+3)+1);
+			*p = '\t';
+		}
+		if (p || (p = strchr(gopherpath, '\t'))) {
+			asprintf(&entry->tag, "%s", gopherpath);
+			*p = '\0';
+		}
+	}
 	entry->host = host;
 	entry->port = port;
 	entry->entry = entry;
@@ -863,13 +968,19 @@ setup(void)
 	close(fd);
 	if ((devnullfd = open("/dev/null", O_WRONLY)) < 0)
 		die("open: /dev/null: %s", strerror(errno));
-	if (mkdir(tmpdir, S_IRWXU) < 0 && errno != EEXIST)
+
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART;
+	sa.sa_handler = exit;
+	sigaction(SIGINT, &sa, NULL);
+	sigaction(SIGHUP, &sa, NULL);
+	sigaction(SIGTERM, &sa, NULL);
+
+	if (!mkdtemp(tmpdir))
 		die("mkdir: %s: %s", tmpdir, strerror(errno));
 	if(interactive = isatty(1)) {
 		uisetup();
-		sigemptyset(&sa.sa_mask);
 		sa.sa_handler = uisigwinch;
-		sa.sa_flags = SA_RESTART;
 		sigaction(SIGWINCH, &sa, NULL);
 	}
 }
@@ -885,10 +996,13 @@ main(int argc, char *argv[])
 	mainurl = xstrdup(argv[1]);
 
 	mainentry = moldentry(mainurl);
-	if (interactive)
+	if (interactive) {
+		diag = uistatus;
 		delve(mainentry);
-	else
+	} else {
+		diag = stddiag;
 		printout(mainentry);
+	}
 
 	exit(0);
 }
